"""This is a template package to demonstrate Continuous Integration."""

from .field import Array, Category, Field, Identifier, MultiArray, Variable
from .taxdump import Taxdump
from .dataset import Metadata
